/* LpEntry.cpp
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <malloc.h>
#include "LpEntry.h"

static const char*	KEY_TAG		= "key";
static const char*	VALUE_TAG	= "value";

/*************************************************************************
 * LP-ENTRY
 *************************************************************************/
LpEntry::LpEntry()
		: mName(0), mUrl(0)
{
}

LpEntry::LpEntry(const BMessage *entryMsg)
		: mName(0), mUrl(0)
{
	if (entryMsg != 0) FromMessage(entryMsg);
}

LpEntry::~LpEntry()
{
	FreeMemory();
}

const char* LpEntry::Name() const
{
	return (const char*)mName;
}

void LpEntry::SetName(const char* name)
{
	if (mName != 0) free(mName);
	mName = 0;
	if (name == 0) return;

	mName = (char*)malloc( sizeof(char) * (strlen(name) + 5) );
	if (mName != 0) strcpy(mName, name);
}

const char* LpEntry::Url() const
{
	return (const char*)mUrl;
}

void LpEntry::SetUrl(const char* url)
{
	if (mUrl != 0) free(mUrl);
	mUrl = 0;
	if (url == 0) return;

	mUrl = (char*)malloc( sizeof(char) * (strlen(url) + 5) );
	if (mUrl != 0) strcpy(mUrl, url);
}

status_t LpEntry::KeyValueAt(uint32 index,
							const char** key,
							const char** value) const
{
	if ( index >= mKeys.size() ) return B_ERROR;
	if ( index >= mValues.size() ) return B_ERROR;
	*key = mKeys[index];
	*value = mValues[index];
	return B_OK;
}

status_t LpEntry::AddKeyValue(	const char* key,
								const char* value)
{
	if ( (key == 0) || (value == 0) ) return B_ERROR;
	char*	keyStr = 0;
	char*	valStr = 0;

	keyStr = (char*)malloc( sizeof(char) * (strlen(key) + 5) );
	if (keyStr == 0) return B_NO_MEMORY;
	valStr = (char*)malloc( sizeof(char) * (strlen(value) + 5) );
	if (valStr == 0) {
		free(keyStr);
		return B_NO_MEMORY;
	}
	strcpy(keyStr, key);
	strcpy(valStr, value);
	mKeys.push_back(keyStr);
	mValues.push_back(valStr);
	return B_OK;
}

void LpEntry::DeleteKeyValues()
{
	FreeVector(mKeys);
	FreeVector(mValues);
}

BMessage* LpEntry::AsMessage() const
{
	BMessage*	msg =  new BMessage(ARPMSG_ENTRY);
	if (msg == 0) return 0;
	if (mName != 0) msg->AddString(NAME_TAG(), mName);	
	if (mUrl != 0) msg->AddString(URL_TAG(), mUrl);	

	const char*	key;
	const char*	value;
	for (uint32 k=0; KeyValueAt(k, &key, &value) == B_OK; k++) {
		msg->AddString(KEY_TAG, key);
		msg->AddString(VALUE_TAG, value);
	}
	return msg;
}

ssize_t LpEntry::WriteTo(LpXmlStreamI& stream) const
{
	if (Name() == 0) return B_ERROR;
	ssize_t		answer, total = 0;
	// Write <Entry name = 'name' url = 'url'>
	string_vec	keys, values;
	keys.push_back( (char*)NAME_TAG() );
	values.push_back( (char*)Name() );
	if (Url() != 0) {
		keys.push_back( (char*)URL_TAG() );
		values.push_back( (char*)Url() );
	}
	if ((answer = stream.WriteStartTag(	XML_TAG(),
										&keys,
										&values)) <= 0)
		return answer;
	total += answer;
	stream.WriteCR();
	stream.PushTab();

	for (uint32 k=0; k<mKeys.size(); k++) {
		if ( k<mValues.size() ) {
			string_vec		keys, values;
			keys.push_back( (char*)KEY_TAG );
			values.push_back(mKeys[k]);
			keys.push_back( (char*)VALUE_TAG );
			values.push_back(mValues[k]);
			if ((answer = stream.WriteStartTag("key_value", &keys, &values)) <= 0) return answer;
			total += answer;
			if ((answer = stream.WriteEndTag("key_value")) <= 0) return answer;
			total += answer;
		}
	}

	stream.PopTab();
	// Write </Entry?
	stream.WriteTabs();
	if ((answer = stream.WriteEndTag(XML_TAG())) <= 0) return answer;
	return total + answer;
}

ssize_t LpEntry::ReadFrom(LpXmlStreamI& stream)
{
	ssize_t		answer = 0;
	string_vec	keys, values;
	char		*tag = stream.ReadNextTag(&keys, &values);
	if (tag == 0) return 0;
	while ( (tag != 0)
			&& (strcmp(tag, XML_END_TAG() ) != 0) ) {
		if (strcmp(tag, "key_value") == 0) {
			ReadKeyValue(keys, values);
		}
		free(tag);
		FreeVector(keys);
		FreeVector(values);
		tag = stream.ReadNextTag(&keys, &values);
	}
	return answer;
}

void LpEntry::ReadKeyValue(	string_vec& keys,
							string_vec& values)
{
	if ( (keys.size() < 2) || (values.size() < 2) ) return;
	if ( strcmp(keys[0], KEY_TAG) != 0 ) return;
	if ( strcmp(keys[1], VALUE_TAG) != 0 ) return;
	AddKeyValue(values[0], values[1]);
}

void LpEntry::PrintToStream() const
{
	printf("LpEntry name: %s, URL: %s\n", mName, mUrl); fflush(stdout);
	if ( mKeys.size() != mValues.size() ) {
		printf("\tERROR!  Mismatch in key and value size\n");
		return;
	}
	if (mKeys.size() == 0) {
		printf("\tno keys and values\n");
		return;
	}
	for (uint32 k=0; k<mKeys.size(); k++)
		printf("\tkey: %s value: %s\n", mKeys[k], mValues[k]);
}

bool LpEntry::operator<(const LpEntry* item) const
{
	if (Name() == 0) return false;
	if (item->Name() == 0) return true;
	if (strcmp(Name(), item->Name()) <= 0) return true;
	return false;
}

bool LpEntry::operator==(const LpEntry* item) const
{
	if (Name() == 0) return false;
	if (item->Name() == 0) return false;
	if (strcmp(Name(), item->Name()) == 0) return true;
	return false;
}

bool LpEntry::less_than(LpEntry* x, LpEntry* y)
{
	return *x < y;
}

void LpEntry::FromMessage(const BMessage *entryMsg)
{
	assert(entryMsg != 0);
	if (entryMsg->what != ARPMSG_ENTRY) return;
	SetName(0);
	SetUrl(0);
	const char*		str;
	if (entryMsg->FindString(NAME_TAG(), &str) == B_OK) SetName(str);
	if (entryMsg->FindString(URL_TAG(), &str) == B_OK) SetUrl(str);

	DeleteKeyValues();
	const char*		key;
	const char*		value;
	for (int32 k=0; entryMsg->FindString(KEY_TAG, k, &key) == B_OK; k++) {
		if ( entryMsg->FindString(VALUE_TAG, k, &value) == B_OK )
			AddKeyValue(key, value);
	}
}

void LpEntry::FreeMemory()
{
	if (mName != 0) free(mName);
	if (mUrl != 0) free(mUrl);
	DeleteKeyValues();
}