/* LpXmlStream.cpp
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "LpXmlStream.h"

#define STR_START_L		"<"
#define STR_START_R		">"
#define STR_END_L		"</"
#define STR_END_R		">"

#define CHR_START_L		'<'
#define CHR_START_R		'>'

/*************************************************************************
 * I-READ-UP-TO
 * A little class to perform the ReadUpTo operation, since it has state
 * and I don't want to muddy up the main reading class with it.
 *************************************************************************/
class IReadUpTo
{
public:
	IReadUpTo(BPositionIO *stream);
	~IReadUpTo();
	
	char* EndString(char *endStr);

private:
	BPositionIO		*mStream;		// Stream being read from
	char*			mBuffer;		// Current read buffer
	ssize_t			mBufferLength;	// Size of read buffer
	ssize_t			mBufferPos;		// Current location in buffer
	const char*		mEndStr;		// String that ends this read
	ssize_t			mEndLength;		// Length of end string
	
	status_t		GrowBuffer(ssize_t req);
};

IReadUpTo::IReadUpTo(BPositionIO *stream)
		: mStream(stream), mBuffer(0), mBufferLength(0), mBufferPos(0),
		  mEndStr(0), mEndLength(0)
{
}

IReadUpTo::~IReadUpTo()
{
	free(mBuffer);
}

status_t IReadUpTo::GrowBuffer(ssize_t req)
{
	if( !mBufferLength ) {
		mBufferLength = 256;
		mBuffer = (char*)malloc(mBufferLength);
	} else if( req > mBufferLength ) {
		mBufferLength = mBufferLength*2;
		if( mBufferLength < req ) mBufferLength = req*2;
		mBuffer = (char*)realloc(mBuffer, mBufferLength);
	}
	
	return mBuffer ? B_OK : B_NO_MEMORY;
}

char* IReadUpTo::EndString(char *endStr)
{
	assert(endStr != 0);
	mEndStr = endStr;
	mEndLength = strlen(endStr);
	mBufferPos = 0;
	
	while( 1 ) {
		if( GrowBuffer(mBufferPos + mEndLength) != B_OK ) return 0;
		
		// Read next batch of bytes.
		if( mStream->Read(&mBuffer[mBufferPos], mEndLength) < B_OK ) return 0;
		
		// See if there is a match
		ssize_t p = mBufferPos-mEndLength;
		mBufferPos += mEndLength;
		while( p < (mBufferPos-mEndLength) ) {
			if( p >= 0 && mBuffer[p] == *mEndStr ) {
				if( strncmp(&mBuffer[p], mEndStr, mEndLength) == 0 ) {
					// The end!
					char* ret = (char*)malloc(p+1);
					memcpy(ret, &mBuffer[0], p);
					ret[p] = 0;
					if( p != mBufferPos ) {
						mStream->Seek(p-mBufferPos+1, SEEK_CUR);
					}
					mBuffer[mBufferPos-1] = 0;
					fflush(stdout);
					return ret;
				}
			}
			p++;
		}
	}
}

/*************************************************************************
 * EX-CHAR-RULE CLASSES
 *************************************************************************/
// This block answers true for any of my white space characters.
class IWsRules : public ExCharRules
{
public:
	IWsRules()					{ }
	virtual bool Accept(char c)
	{
		if (c == 9) return true;	// Tab
		if (c == 13) return true;	// CR
		if (c == 32) return true;	// Space
		return false;
	}
};

// This is a special block for skipping past the separator between
// a key-value pair:  It skips past all white space, plus the first
// occurrence of '='.  IT NEEDS TO BE INITIALIZED.
class IKeyValueRules : public ExCharRules
{
public:
	IKeyValueRules()			{ }
	virtual void Initialize()	{ mPassedEquals = false; }
	virtual bool Accept(char c)
	{
		if (c == 9) return true;	// Tab
		if (c == 13) return true;	// CR
		if (c == 32) return true;	// Space
		if (c == 61) {
			if (mPassedEquals) return false;
			mPassedEquals = true;
			return true;
		}
		return false;
	}
private:
	bool	mPassedEquals;
};

// These rules take a single character and are just based on that.
class ISingleCharRules : public ExCharRules
{
public:
	ISingleCharRules(char c)			{ mC = c; }
	virtual bool Accept(char c)
	{
		return mC == c;
	}
private:
	char	mC;
};

/*************************************************************************
 * LP-XML-STREAM
 *************************************************************************/
LpXmlStream::LpXmlStream(BPositionIO *stream)
		: mTabs(0), mStream(stream)
{
	mBuffer = new IReadUpTo(mStream);
	mWhiteSpace = new IWsRules();
}

LpXmlStream::~LpXmlStream()
{
	delete mBuffer;
	if (mWhiteSpace != 0) delete mWhiteSpace;
}

void LpXmlStream::PushTab()
{
	mTabs++;
}

void LpXmlStream::PopTab()
{
	mTabs--;
	if (mTabs < 0) mTabs = 0;
}

ssize_t LpXmlStream::WriteTabs()
{
	ssize_t		answer, total = 0;
	char	str[3];
	str[0] = 9;
	str[1] = 0;

	for (int16 k=0; k<mTabs; k++) {
		if ( (answer = mStream->Write( (void*)str, sizeof(char) * strlen(str))) <= 0)
			return answer;
		total += answer;
	}
	return total;
}

ssize_t LpXmlStream::WriteChar(char c)
{
	char	str[3];
	str[0] = c;
	str[1] = 0;
	return mStream->Write( (void*)str, sizeof(char) * strlen(str));
}

ssize_t LpXmlStream::WriteCR()
{
	return WriteChar(10);
}

ssize_t LpXmlStream::WriteSpace()
{
	return WriteChar(32);
}

ssize_t LpXmlStream::ReadBlanks()
{
	ssize_t		answer, total = 0;
	int8		peek;
	answer = Peek(&peek);
	total += answer;
	while ( (answer == 1)
			&& ( (peek == 0)	// null
				|| (peek == 9)	// tab
				|| (peek == 10)	// line feed
				|| (peek == 11)	// home
				|| (peek == 12)	// form feed
				|| (peek == 13)	// carriage return
				|| (peek == 32) // space
			)) {
		answer = Peek(&peek);
		total += answer;
	}
	return total;
}

ssize_t LpXmlStream::WriteStartTag(	const char *tag,
									string_vec *keys,
									string_vec *values)
{
	WriteTabs();
	ssize_t		answer, total = 0;
	if ( (answer = mStream->Write( (void*)STR_START_L, sizeof(char) * strlen(STR_START_L))) <= 0)
		return answer;
	total = answer;
	if ( (answer = mStream->Write( (void*)tag, sizeof(char) * strlen(tag))) <= 0)
		return answer;
	total += answer;

	if ( (answer = WriteKeysAndValues(keys, values)) < 0)
		return answer;
	total += answer;

	if ( (answer = mStream->Write( (void*)STR_START_R, sizeof(char) * strlen(STR_START_R))) <= 0)
		return answer;
	return total + answer;
}

ssize_t LpXmlStream::Write(const void *buffer, size_t size)
{
	assert(mStream != 0);
	return mStream->Write(buffer, size);
}

ssize_t LpXmlStream::WriteEndTag(const char *tag)
{
	ssize_t		answer, total = 0;
	if ( (answer = mStream->Write( (void*)STR_END_L, sizeof(char) * strlen(STR_END_L))) <= 0)
		return answer;
	total = answer;
	if ( (answer = mStream->Write( (void*)tag, sizeof(char) * strlen(tag))) <= 0)
		return answer;
	total += answer;
	if ( (answer = mStream->Write( (void*)STR_END_R, sizeof(char) * strlen(STR_END_R))) <= 0)
		return answer;
	total += answer;
	if ((answer = WriteCR()) <= 0) return answer;
	return total + answer;
}

char* LpXmlStream::ReadStrToTag(const char *endTag)
{
	char	end[strlen(endTag) + 4];
	sprintf(end, "%s%s%s", STR_END_L, endTag, STR_END_R);
	return ReadUpTo(end);
}

char* LpXmlStream::ReadNextTag(	string_vec *keys,
								string_vec *values)
{
	char	*str;
	if (FindChar(CHR_START_L) != B_OK) return 0;
	str = ReadUpTo(STR_START_R);
	if (str == 0) return 0;
	str = FormatTag(str, keys, values);
	return str;
}

status_t LpXmlStream::FindChar(char data)
{
	char	val;
	ssize_t	answer;
	for (answer = mStream->Read((void*)&val, 1);
			answer == 1;
			answer = mStream->Read((void*)&val, 1)) {
		if (val == data) return B_OK;
	}
	return B_ERROR;
}

ssize_t LpXmlStream::WriteKeysAndValues(string_vec *keys,
										string_vec *values)
{
	assert(mStream != 0);
	if ( (keys == 0) || (values == 0) ) return 0;
	if ( keys->size() < 1) return 0;
	if ( keys->size() != values->size() ) return 0;

	ssize_t		answer, total = 0;
	// This is the string that separates a key from its value.
	char		sep[5];
	sep[0] = 32;
	sep[1] = 61;
	sep[2] = 32;
	sep[3] = 0;
	size_t		sepSize = sizeof(char) * strlen(sep);
	// Write the keys and values.
	for (uint32 k = 0; k < keys->size(); k++) {
		// Write a space separating this key / value from the previous data.
		if ( (answer = WriteSpace()) <= 0)
			return B_ERROR;
		total += answer;

		char	*key = (*keys)[k], *val = (*values)[k];
		size_t	kSize = sizeof(char) * strlen(key);
		if ( (answer = mStream->Write( (void*)key, kSize)) <= 0)
			return B_ERROR;
		total += answer;
		if ( (answer = mStream->Write( (void*)sep, sepSize)) <= 0)
			return B_ERROR;
		total += answer;
		if ( (answer = WriteValue(val)) <= 0)
			return B_ERROR;
		total += answer;
	}
	return total;
}

ssize_t LpXmlStream::WriteValue(char *value)
{
	if (value == 0) return 0;
	size_t	len = sizeof(char) * strlen(value);
	if (Contains(value, len, mWhiteSpace)) {
		ssize_t		answer, total = 0;

		if ( (answer = WriteChar(34)) <= 0)
			return answer;
		total += answer;
		if ( (answer = mStream->Write( (void*)value, len)) <= 0)
			return answer;
		total += answer;
		if ( (answer = WriteChar(34)) <= 0)
			return answer;
		total += answer;
		return total;
	}
	return mStream->Write( (void*)value, len);
}

char* LpXmlStream::FormatTag(	char *fullTag,
								string_vec *keys,
								string_vec *values)
{
	if (mWhiteSpace == 0) {
		free(fullTag);
		return 0;
	}
	size_t	len = strlen(fullTag);
	// If the tag has no whitespace, don't bother just making
	// a copy of it.
	if (Contains(fullTag, len, mWhiteSpace) == false) return fullTag;

	size_t	index = 0;
	char 	*newTag = UpTo(fullTag, len, 0, &index, mWhiteSpace);
	// FIX: an error condition, but how to deal with it?
	if (newTag == 0) return fullTag;

	if ( (keys == 0) || (values == 0) ) {
		free(fullTag);
		return newTag;
	}
	ReadKeysAndValues(fullTag, len, index, keys, values);
	free(fullTag);
	return newTag;
}

void LpXmlStream::ReadKeysAndValues(char *str,
									size_t len,
									size_t index,
									string_vec *keys,
									string_vec *values)
{
	size_t	startI = index,
			endI = index;
	// In case someone doesn't actually read any data, this little
	// var stops us from falling into an infinite loop.
	size_t	stopLoops;

	while (endI < len) {
		char	*key, *value;
		stopLoops = startI;
		startI = endI;
		key = ReadKey(str, len, startI, &endI);
		if (endI >= len) return;

		startI = endI;
		value = ReadValue(str, len, startI, &endI);

		if ( (key != 0) && (value != 0) ) {
			keys->push_back(key);
			values->push_back(value);
		}
		if (endI >= len) return;
		if (startI == stopLoops) return;
	}
}

char* LpXmlStream::ReadKey(	char *str,
							size_t len,
							size_t start,
							size_t *end)
{
	Skip(str, len, start, end, mWhiteSpace);
	if (*end >= len) return 0;
	
	start = *end;
	return UpTo(str, len, start, end, mWhiteSpace);
}

char* LpXmlStream::ReadValue(	char *str,
								size_t len,
								size_t start,
								size_t *end)
{
	IKeyValueRules	rules;
	rules.Initialize();

	Skip(str, len, start, end, &rules);
	if (*end >= len) return 0;

	start = *end;
	// If this value is in quotes, strip them out.
	if (str[start] == 34) {
		start++;
		ISingleCharRules	single(34);
		char		*answer = UpTo(str, len, start, end, &single);
		(*end)++;
		return answer;
	}
	return UpTo(str, len, start, end, mWhiteSpace);
}

char* LpXmlStream::UpTo(char *str,
						size_t len,
						size_t startIndex,
						size_t *stopIndex,
						ExCharRules *rules)
{
	if (startIndex >= len) {
		return 0;
		*stopIndex = len;
	}
	*stopIndex = startIndex;
	while ( ( ((*stopIndex) + 1) < len)
			&& (!rules->Accept(str[(*stopIndex) + 1])) )
		(*stopIndex)++;

	if (*stopIndex > len) return 0;
	size_t	size = *stopIndex - startIndex;
	if (size < 1) return 0;
	char	*copy = (char*)malloc(sizeof(char) * (size + 4));
	if (copy == 0) return 0;
	size_t	k, index = 0;
	for (k=startIndex; k<=(*stopIndex); k++) {
		copy[index] = str[k];
		index++;
	}
	copy[index] = 0;
	(*stopIndex)++;
	return copy;
}

void LpXmlStream::Skip(	char *str,
						size_t len,
						size_t startIndex,
						size_t *stopIndex,
						ExCharRules *rules)
{
	*stopIndex = startIndex;
	while( rules->Accept(str[*stopIndex]) ) {
		(*stopIndex)++;
		if (*stopIndex >= len) return;
	}
}

bool LpXmlStream::Contains(	char *str,
							size_t len,
							ExCharRules *rules)
{
	for (size_t k=0; k<len; k++) {
		if (rules->Accept(str[k])) return true;
	}
	return false;
}

ssize_t LpXmlStream::Peek(int8 *data)
{
	assert(mStream != 0);
	ssize_t		answer;
	off_t		pos = mStream->Position();
	answer = mStream->Read((void*)data, 1);
	mStream->Seek(SEEK_SET, pos);
	return answer;
}

char* LpXmlStream::ReadUpTo(char *endStr)
{
	return mBuffer->EndString(endStr);
}

char* LpXmlStream::GrowBy(char *str, int32 size)
{
	size_t	len = strlen(str);
	char	*newStr = (char*)malloc(sizeof(char) * (len + size + 2));
	if (newStr == 0) return 0;
	for (size_t k=0; k<len; k++) newStr[k] = str[k];
	newStr[len] = '\0';
	free(str);
	return newStr;
}
