/* LpXmlStream.h
 * Copyright (c)2000 by Eric Hackborn.
 * All rights reserved.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Eric Hackborn,
 * at <hackborn@angryredplanet.com>.
 *
 * ----------------------------------------------------------------------
 *
 * Known Bugs
 * ~~~~~~~~~~
 *
 *	- None!  Ha ha!
 *
 * ----------------------------------------------------------------------
 *
 * History
 * ~~~~~~~
 * 04.21.00		hackborn
 * Created this file.
 */

#ifndef LP_LPXMLSTREAM_H
#define LP_LPXMLSTREAM_H

#include <be/support/DataIO.h>
#include "LpXmlStreamI.h"

// Forward references
class IReadUpTo;
class ExCharRules;

/***************************************************************************
 * EX-XML-STREAM
 * This class defines the behaviour necessary for ExFiler classes to write
 * to and read from an XML-formatted stream.
 ***************************************************************************/
class LpXmlStream : public LpXmlStreamI
{
public:
	/* If a file size is supplied, it can be used for percentage feedback.
	 */
	LpXmlStream(BPositionIO *stream);
	virtual ~LpXmlStream();

	virtual VersionType Version(void) const			{ return VERSION_CURRENT; }
	
	virtual void PushTab();
	virtual void PopTab();
	/* Write a tab character for however many tabs have been
	 * pushed on me.
	 */
	virtual ssize_t WriteTabs();
	virtual ssize_t WriteChar(char c);
	virtual ssize_t WriteCR();
	virtual ssize_t WriteSpace();
	/* Read past blank characters -- tabs, spaces, carriage returns, etc.
	 */
	ssize_t ReadBlanks();

	/* Each Write() function answers the number of bytes written,
	 * or a negative error value.
	 */	
	virtual ssize_t WriteStartTag(	const char *tag,
									string_vec *keys = 0,
									string_vec *values = 0);
	virtual ssize_t Write(const void *buffer, size_t size); 
	virtual ssize_t WriteEndTag(const char *tag);

	/* Answer the next tag in the stream.  The client owns the data
	 * returned, so they must dispose of it.
	 */
	virtual char* ReadNextTag(	string_vec *keys = 0,
								string_vec *values = 0);
	virtual char* ReadStrToTag(const char *endTag);
	
	/* Answer B_OK if the tag was read and the stream is positioned
	 * immediately after it.
	 */
//	status_t ReadStartTag(const char *tag);

protected:
	/* A rule block that defines my separator characters.
	 */
	ExCharRules		*mWhiteSpace;
	/* Read through the stream looking for the character.  If it isn't
	 * found, answer B_ERROR, otherwise B_OK.  The stream ends up
	 * positioned immediately after the character, or at the end if it
	 * wasn't found.
	 */
	status_t FindChar(char data);

	/* Write the keys and values in the form of "[key] = [value]".
	 * The keys and values vectors must be the same size.  Answer
	 * the number of bytes written, or 0 if keys and values weren't
	 * valid vectors, or a negative value for an error.
	 */
	ssize_t WriteKeysAndValues(	string_vec *keys,
								string_vec *values);
	/* Write out the value.  If the value has any whitespace, automatically
	 * put quotes around it.
	 */
	ssize_t WriteValue(char *value);
	/* This method takes a read in tag, fullTag, and chops it up
	 * into a new tag (the answered char*) and any keys and values
	 * that might be in it.  It works like this:  The fullTag is
	 * truncated at its first separator character (space, tab, etc),
	 * and that truncated value becomes the new tag.  Then key value
	 * pairs are read in, in the form '[key] = [value]' and placed
	 * in the vectors (unless the vectors are null, and then the
	 * keys and values are just free'd.
	 */
	char* FormatTag(char *fullTag,
					string_vec *keys,
					string_vec *values);
	void ReadKeysAndValues(	char *str,
							size_t len,
							size_t index,
							string_vec *keys,
							string_vec *values);
	char* ReadKey(	char *str,
					size_t len,
					size_t start,
					size_t *end);
	char* ReadValue(char *str,
					size_t len,
					size_t start,
					size_t *end);

	char* UpTo(	char *str,
				size_t len,
				size_t startIndex,
				size_t *stopIndex,
				ExCharRules *rules);
	void Skip(	char *str,
				size_t len,
				size_t startIndex,
				size_t *stopIndex,
				ExCharRules *rules);
	bool Contains(	char *str,
					size_t len,
					ExCharRules *rules);
	
private:
	// The number of tabs that I write out before each line.
	int16			mTabs;

	// The stream I read / write to.
	BPositionIO		*mStream;
	
	/* Answer the next byte of data in the stream
	 */
	ssize_t Peek(int8 *data);

	/* State for current file read
	 */
	IReadUpTo* mBuffer;
	
	/* Answer a string starting with the current stream position and
	 * ending with the first character before str.  Answer 0 if str
	 * does not occur before the end of the stream, otherwise answer
	 * a newly allocated string that is the responsibility of the
	 * client.  The stream ends up positioned after str or at the end
	 * of the stream, whichever comes first.
	 */
	char* ReadUpTo(char *endStr);
	/* Take character array str, copy it into a new array that's size
	 * characters larger, delete str, and answer the new array.  Answerr
	 * 0 if there's not enough memory.
	 */
	char* GrowBy(char *str, int32 size);
};

/*************************************************************************
 * EX-CHAR-RULES
 * This class defines rules for examining characters.  It will answer
 * true or false based on the character supplied and its own internal
 * state.  Note that because some rule blocks maintain state, you need
 * to be careful to make sure they're initialized.
 *************************************************************************/
class ExCharRules
{
public:
	virtual ~ExCharRules()		{ }
	virtual void Initialize()	{ }
	virtual bool Accept(char c) = 0;
};

#endif
